/*
 * Copyright (c) 2015 Stefan Scheffler
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * screen grabber based on glc http://nullkey.ath.cx/projects/glc
 *
 */

/*
#include <libgen.h>
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include <X.h>
#include <GL/glx.h>

#define BASE_SERVER_PORT 57725
#define MAX_SERVERS 20

/* color of gate icon */
#define GATE_R 3
#define GATE_G 192
#define GATE_B 230
#define GATE_TOLERANCE 50

/* color of exit point icon */
#define EXIT_R 0
#define EXIT_G 255
#define EXIT_B 0
#define EXIT_TOLERANCE 50

/* color of target circle */
#define TARGET_R 240
#define TARGET_G 0
#define TARGET_B 0
#define TARGET_TOLERANCE 50

/* object types */
#define OT_GATE 1
#define OT_EXIT 2
#define OT_TARGET 3
#define OT_DOCK 4
#define OT_DOCKLIGHT1 5
#define OT_DOCKLIGHT2 6
#define OT_DOCKNOFOCUS 7
#define OT_TARGETTRACK 8

/* request type */
#define RT_TYPE 1
#define RT_QUERY 2
#define RT_INPUT 3
#define RT_BLOCKWARP 4

typedef struct frame {
	unsigned char *buf;
	unsigned int width;
	unsigned int height;
} frame_t;

/* shm segment struct */
typedef struct shared {
	/* reply */
	char found;
	unsigned int x;
	unsigned int y;

	/* request */
	int process;
	unsigned int offsetx;
	unsigned int offsety;
	unsigned int sizex;
	unsigned int sizey;

	/* object type */
	unsigned char r;
	unsigned char g;
	unsigned char b;
	unsigned char tolerance;

	/* input */
	int input;
	unsigned int inputx, inputy;

	int blockwarp;
} shared_t;

typedef void (*glXSwapBuffers_t)(Display *, GLXDrawable drawable);
typedef int (*XWarpPointer_t)(Display *display,
			      Window src_w,
			      Window dest_w,
			      int src_x,
			      int src_y,
			      unsigned int src_width,
			      unsigned int src_height,
			      int dest_x,
			      int dest_y);

/* globals */
glXSwapBuffers_t glXSwapBuffers_org;
XWarpPointer_t XWarpPointer_org;
shared_t *shared;


int find_icon(frame_t *frame, unsigned int *posx, unsigned int *posy) {
	int x, y;
  
	int offsety_flipped = frame->height - shared->offsety - shared->sizey - 1;

	for (y = offsety_flipped + shared->sizey; y > offsety_flipped ; y--) {
		for (x = shared->offsetx; x <= shared->offsetx + shared->sizex; x++) {
			int offset = (y * frame->width + x) * 3;

			if ((abs(frame->buf[offset    ] - shared->r) <= shared->tolerance) &&
			    (abs(frame->buf[offset + 1] - shared->g) <= shared->tolerance) &&
			    (abs(frame->buf[offset + 2] - shared->b) <= shared->tolerance)) {

				*posx = x;
				*posy = frame->height - y - 1;
				return 1;
			}
		}
	}
	return 0;
}
  
int frame_init(frame_t *frame, unsigned char *buf, unsigned int width, unsigned int height) {
	if ((width < 0) || (height < 0)) {
		fprintf(stderr, "frame dimensions out of range %ix%i", width, height);
	}
	frame->buf = buf;
	frame->width = width;
	frame->height = height;
	return 1;
}

void send_key(Display *dpy, Window drawable) {
	XMotionEvent ev;
	Window root;

	unsigned int width;
	unsigned int height;
	int unused;

	
	XGetGeometry(dpy, drawable, &root, &unused, &unused, &width, &height, (unsigned int *)&unused, (unsigned int *)&unused);

	ev.type = MotionNotify;
	ev.display     = dpy;
	ev.window      = drawable;
	ev.root        = root;
	ev.subwindow   = None;
	ev.time        = CurrentTime;
	ev.x           = width / 2;
	ev.y           = height / 2;
	ev.x_root      = 1;
	ev.y_root      = 1;
	ev.state       = 0;
	ev.is_hint     = NotifyNormal;
	ev.same_screen = True;
    
	XSendEvent(dpy, drawable, True, PointerMotionMask, (XEvent *)&ev);
	XSync(dpy, False);

	ev.type = MotionNotify;
	ev.time        = CurrentTime;
	ev.x           = shared->inputx;
	ev.y           = shared->inputy;
    
	XSendEvent(dpy, drawable, True, PointerMotionMask, (XEvent *)&ev);
	XSync(dpy, False);
}


int XWarpPointer(Display *display,
	     Window src_w,
	     Window dest_w,
	     int src_x,
	     int src_y,
	     unsigned int src_width,
	     unsigned int src_height,
	     int dest_x,
	     int dest_y) {
	
	if (shared->blockwarp) {
		fprintf(stderr, "blocking warp\n");
		return;
	}
	return XWarpPointer_org(display, src_w, dest_w, src_x, src_y, src_width, src_height, dest_x, dest_y);
}


#define NUMKEYS 4
int keys[] = {
	XK_Left,
	XK_Right,
	XK_Up,
	XK_Down
};

void glXSwapBuffers(Display *dpy, GLXDrawable drawable) {
	unsigned int i;
	unsigned int width;
	unsigned int height;
	Window root;
	int unused;
	unsigned char *buf;
	frame_t frame;
	unsigned int x, y;

	/*
	  unsigned char wdata[] = {[0 ... 15] = 255};
	*/
	
	if (shared->input) {
		send_key(dpy, drawable);
		shared->input = 0;
	}
	
	if (shared->process) {
		/* from glc */
		/* get the dimensions of the glx drawable */
		XGetGeometry(dpy, drawable, &root, &unused, &unused, &width, &height, (unsigned int *)&unused, (unsigned int *)&unused);

		buf = malloc(width * height * 3 * sizeof(char));
		if (!buf) {
			shared->found = 0;
			shared->process = 0;
			perror("glXSwapBuffers");
			glXSwapBuffers_org(dpy, drawable);
			return;
		}

		/* get frame */
		glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, buf);
		glPopClientAttrib();
    
		if (!frame_init(&frame, buf, width, height)) {
			free(buf);
			shared->found = 0;
			shared->process = 0;
			glXSwapBuffers_org(dpy, drawable);
			return;
		}

		if (find_icon(&frame, &x, &y)) {
			shared->x = x;
			shared->y = y;
			shared->found = 1;
		}
		else {
			shared->found = 0;
		}

		free(buf);
		shared->process = 0;
	}
	/*
	glWindowPos2i(shared->x, height - shared->y);
	glDrawPixels(4, 4, GL_GREEN, GL_UNSIGNED_BYTE, wdata);
	*/
	glXSwapBuffers_org(dpy, drawable);
}


int have_linebreak(char *buf, int len) {
	int i;

	for (i = 0; i < len; i++) {
		if (buf[i] == '\n') {
			return i;
		} 
	}

	return -1;
}


int get_param(char **buf) {
	char param[20];
	char *buf_ptr = *buf;
	char *end_ptr;
	int ret;
	int i;

	for (i = 0; i < 9; i++) {
		char chr = *buf_ptr;

		buf_ptr++;
		if ((chr == ' ') || (chr == '\0')) {
			break;
		}

		param[i] = chr;
	}

	param[i] = '\0';
	*buf = buf_ptr;

	//	errno = 0;
	ret = (int)strtol(param, &end_ptr, 10);

	if (ret == LONG_MAX || ret == LONG_MIN) {
		perror("get_param");
		ret = -1;
	} 
	else if (param == end_ptr) {
		fprintf(stderr, "get_param: not a number\n");
		ret = -1;
	}
  
	return ret;
}


void send_reply(int csock, char *buf) {
	char *send_ptr = buf;
	int left;
	
	left = strlen(buf);
	fprintf(stderr, "send_reply %i %s", left, buf);
	while (left > 0) {
		int sent = send(csock, send_ptr, left, 0);
		send_ptr+=sent;
		left-=sent;
	}
}

void handle_msg(int csock, char *msg) {
	char sendbuf[30];

       	switch (get_param(&msg)) {
	case RT_TYPE:
		send_reply(csock, "1\n");
		break;
	case RT_QUERY:
		/* TODO: this should go into a seperate function */
		if (((shared->offsetx = get_param(&msg)) < 0) ||
		    ((shared->offsety = get_param(&msg)) < 0) ||
		    ((shared->sizex = get_param(&msg)) < 0) ||
		    ((shared->sizey = get_param(&msg)) < 0)) {
	
			fprintf(stderr, "parameter error\n");
			close(csock);
			return;
		} 
      
		switch(get_param(&msg)) {
		case OT_GATE:
			shared->r = GATE_R;
			shared->g = GATE_G;
			shared->b = GATE_B;
			shared->tolerance = GATE_TOLERANCE;
			break;
		case OT_EXIT:
			shared->r = EXIT_R;
			shared->g = EXIT_G;
			shared->b = EXIT_B;
			shared->tolerance = EXIT_TOLERANCE;
			break;
		case OT_TARGET:
		case OT_DOCK:
		case OT_DOCKLIGHT1:
		case OT_DOCKLIGHT2:
		case OT_DOCKNOFOCUS:
		case OT_TARGETTRACK:
			shared->r = TARGET_R;
			shared->g = TARGET_G;
			shared->b = TARGET_B;
			shared->tolerance = TARGET_TOLERANCE;
			break;
		default:
			fprintf(stderr, "unknown object type\n");
			close(csock);
			return;
		}
      
		shared->process = 1;
      
		/* TODO: could probably use signaling for this */
		while (1) {
			if (!shared->process) {
				break;
			}
			usleep(100000);
		}      
      
		if (shared->found) { 
			snprintf(sendbuf, 30, "%i %i\n", shared->x, shared->y);
			fprintf(stderr, sendbuf);
			send_reply(csock, sendbuf);
		}
		else {
			send_reply(csock, "-1 -1\n");
		}
		break;
	case RT_INPUT:
		fprintf(stderr, "input\n");
		shared->inputx = get_param(&msg);
		shared->inputy = get_param(&msg);
		shared->input = 1;
		break;
	case RT_BLOCKWARP:
		shared->blockwarp = get_param(&msg);
		fprintf(stderr, "block %i\n", shared->blockwarp);
		break;
	}

}

int server() {
	int sock;
	unsigned int server_index = 0;
	char *server_index_p;
	
	struct sockaddr_in addr, caddr;
	socklen_t caddr_len = sizeof(caddr);

	
	server_index_p = getenv("ROBOTSERVER_INDEX");
	if (server_index_p) {
		server_index = atoi(server_index_p);
		if (server_index > MAX_SERVERS) {
			fprintf(stderr, "invalid ROBOTSERVER_INDEX\n");
			return 0;
		}
	}
	fprintf(stderr, "server starting\n");
  
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("server socket");
		return 0;
	}
  
	addr.sin_family = AF_INET;
	addr.sin_port = htons(BASE_SERVER_PORT + server_index);
	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	if (addr.sin_addr.s_addr == -1) {
		fprintf(stderr, "inet_addr error\n");
		return 0;
	}

	memset(addr.sin_zero, 0, sizeof(addr.sin_zero));
  
	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
		perror("server bind");
		return 0;
	}

	if (listen(sock, 5) == -1) {
		perror("server listen");
		return 0;
	}
  
	while (1) {
		int csock;
		int len;
    
		char buf[1024];
		char *buf_ptr = buf;
		int buf_left = 1023;
		int reply_len = 0;
    
		if ((csock = accept(sock, (struct sockaddr *)&caddr, &caddr_len)) == -1) {
			perror("server accept");
			continue;
		}
    
		fprintf(stderr, "client connected\n");

		while (1) {
			while ((len = recv(csock, buf_ptr, 1024, 0)) > 0) {
				if (len < 0) {
					break;
				}
				
				buf_ptr+=len;
				buf_left-=len;
				
				if ((reply_len = have_linebreak(buf, 1024 - buf_left)) > -1) {
					buf[reply_len] = '\0';

					/* fprintf(stderr, "handling message %s\n", buf); */
					
					handle_msg(csock, buf);
					
					memmove(buf, buf + (reply_len + 1), 1024 - (reply_len + 1));
					buf_ptr = buf;
					buf_left += (reply_len + 1);
				}
			}
			
			if (len < 0) {
				perror("recv");
				close(csock);
				break;
			}
			else if (len == 0) {
				fprintf(stderr, "client disc\n");
				close(csock);
				break;
			}
		}
	}

	return 1;
}

void sigchld_handler(int arg) {
	exit(0);
}


void __attribute__ ((constructor)) libinit() {
	void *handle;
	char *err;
	int shmid;

	/*
	char *exename;
	char *path;
	
	path = realpath("/proc/self/exe", NULL);
	exename = basename(path);
	fprintf(stderr, "exepath %s\n", exename);
	if (strcmp(exename, "vendetta")) {
		return;
	}
	free(path);
	*/
	
	if (!(shmid = shmget(IPC_PRIVATE, sizeof(shared_t), 0600))) {
		perror("libinit");
		return;
	}

	switch (fork()) {
	case 0:
		shared = shmat(shmid, NULL, 0);

		if (shared == (shared_t *)(-1)) {
			perror("child shmat");
			return;
		}

		/* save original glXSwapBuffers */
		if (!(handle = dlopen("/usr/lib/libGL.so.1", RTLD_LAZY))) {
			fprintf(stderr, "dlopen error %s\n", dlerror());
			return;
		}
    
		dlerror();
		glXSwapBuffers_org = dlsym(handle, "glXSwapBuffers");
    
		if ((err = dlerror())) {
			fprintf(stderr, "symbol lookup error: %s \n", err);
			return;
		}

		/* same for XWarpPointer */
		if (!(handle = dlopen("/usr/lib/libX11.so.6", RTLD_LAZY))) {
			fprintf(stderr, "dlopen error %s\n", dlerror());
			return;
		}
    
		dlerror();
		XWarpPointer_org = dlsym(handle, "XWarpPointer");
    
		if ((err = dlerror())) {
			fprintf(stderr, "symbol lookup error: %s \n", err);
			return;
		}

    
		break;
	case -1:
		shmctl(shmid, IPC_RMID, NULL);
		perror("fork");
		break;
	default:
		signal(SIGCHLD, sigchld_handler);

		shared = shmat(shmid, NULL, 0);
		shmctl(shmid, IPC_RMID, NULL);
    
		if (shared == (shared_t *)(-1)) {
			perror("server shmat");
			return;
		}

		if (!server()) {
			fprintf(stderr, "server died\n");
		}
		if (wait(NULL) == -1) {
			perror("wait");
		}
	}
}
