-- object finder for tspammer

-- Copyright (c) 2015 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.


local finder = {}
local TCP = dofile("tcpsock.lua")

local function roboprint(str)
--	print("objectfinder: "..tostring(str))
end

finder.TYPE_PNG = 0
finder.TYPE_GL = 1

finder.FIND_SUCCESS = 1
finder.FIND_LOST = 2
finder.FIND_CONERROR = 3
finder.FIND_STUCK = 4

-- object types
finder.OT_DEFAULT = 0
finder.OT_TARGETTRACK = 8


function finder.new()
	local instance = {}

	local r_off_x = 0
	local r_off_y = 0
	local r_centerx = 0
	local r_centery = 0
	local r_size = 0

	local blib_correction = 0

	local radarimage
	local radarcolor

	local blindfind_y
	
	local sock
		
	instance.spin_timer = Timer()

	local function get_radar()
		return iup.GetBrother(iup.GetParent(HUD.leftradarlabel))
	end

	local function setup_radar_info()
		local radar = get_radar()

		if iup.GetType(radar) == "radar" then
			r_off_x = tonumber(radar.X)
			r_off_y = tonumber(radar.Y)

			r_size = tonumber(radar.h)

			r_centerx = r_off_x + (r_size / 2)
			r_centery = r_off_y + (r_size / 2)

			blib_correction = math.floor(r_size / 20)
		else
			console_print("Warning! can't get radar info.")
		end
	end



	-- reply handlers for each object type
	local replyhandler = {}

	replyhandler[finder.OT_TARGETTRACK] = function(object_type, cb, x, y)

		local function send_input_request(x, y)
			sock.tcp:Send("3 "..tostring(x).." "..tostring(y).."\n")
		end
		
		local scaled_x = 0
		local scaled_y = 0
		
		if x and x >= 0 and y and y >= 0 then
			scaled_x = (r_centerx - x) / (r_size / 2)
			scaled_y = (r_centery - y) / (r_size / 2)

			local scaled_x2 = (2 * math.asin(scaled_x)) / math.pi
			local scaled_y2 = (2 * math.asin(scaled_y)) / math.pi
			
			if math.abs(scaled_x2) > 0.01 or math.abs(scaled_y2) > 0.01 then
				local xres2 = gkinterface.GetXResolution() / 2
				local yres2 = gkinterface.GetYResolution() / 2

				-- number of pixels to turn 90 deg
				-- moving cursor to the horizonal edge turns the ship by about 15 deg
				-- that is then multiplied with the mouse sensitivity
				local pix_ninety = (6 * yres2) / gkinterface.GetMouseSensitivity()

				-- clamp to yres2
				local pix_x = math.floor(math.min(math.max(pix_ninety * scaled_x2, -yres2), yres2))
				local pix_y = math.floor(math.min(math.max(pix_ninety * scaled_y2, -yres2), yres2))

				send_input_request(xres2 - pix_x, yres2 - pix_y)
			end
		else
			local xres2 = gkinterface.GetXResolution() / 2

			local yres = gkinterface.GetYResolution()
			local ypos = blindfind_y and yres or 0
			blindfind_y = not blindfind_y
			
			send_input_request(xres2, ypos)
		end


		
		local function spin_cb()
			instance:find(object_type, cb)
		end
		instance.spin_timer:SetTimeout(100, spin_cb)
	end


	
	-- common stuff for both finders
	-- does all the networking stuff
	-- calls one of the object type handlers above with the server reply
	function instance:find_common(object_type, cb)		

		local function send_find_request()
			sock.tcp:Send("2 "..
				  tostring(r_off_x).." "..
				  tostring(r_off_y).." "..
				  tostring(r_size-1).." "..
				  tostring(r_size-1).." "..
				  tostring(object_type).."\n")
		end

		local function conn_cb(conn, err)
			if err then
				console_print("Can't connect to dullbot server. error: "..tostring(err))
				sock.tcp:Disconnect()
				sock = nil
				return
			end
			send_find_request()
		end

		local function line_cb(conn, line)

			local _, _, x, y = line:find("([^ ]-) ([^ ]-)$")
			
			x = tonumber(x)
			y = tonumber(y)
			if not x or not y then
				print("Bad reply from dullbot server")
				sock.tcp:Disconnect()
				sock = nil
				if cb then cb(finder.FIND_CONERROR) end
				return
			end

			if x < 0 then x = nil end
			
			if y < 0 then
				y = nil
			else
				y = y + blib_correction
			end

			replyhandler[object_type](object_type, cb, x, y)
		end
		
		local function disc_cb()
			sock.tcp:Disconnect()
			sock = nil
		end

		if sock then
--			print("already connected ..just finding")
			send_find_request()
		else
--			print("connecting...")
			sock = TCP.make_client("127.0.0.1", self.port, conn_cb, line_cb, disc_cb)
		end
	end	

	local function radar_blank()
		local radar = get_radar()
		radar.bgcolor = "0 0 0 255 *"
		radar.IMAGE = ""
		iup.Refresh(radar)
	end
	
	local function radar_restore()
		local radar = get_radar()
		if radarcolor and radarimage then
			radar.bgcolor = radarcolor
			radar.IMAGE = radarimage
		end
	end
	
	instance.find = instance.find_common

	function instance:Init(sdir)
		local radar = get_radar()
		radarimage = radar.IMAGE
		radarcolor = radar.bgcolor

		setup_radar_info()

		radar_blank()
	end

	function instance:Stop()
		self.spin_timer:Kill()
		
		if sock then
			sock.tcp:Disconnect()
			sock = nil
		end
		
		radar_restore()
	end

	function instance:SetPort(port)
		self.port = port
	end

	return instance
end

return finder
